from typing import Union, Optional

from os import PathLike
from pathlib import Path

import numpy as np
from PIL import Image

from readlif.reader import LifFile

from analyzefrc.read import FRCMeasureSettings, FRCMeasurement, FRCSet, frc1_set, frc2_set


__all__ = ['get_image', 'lif_read', 'image_read']


def return_path(pth: str):
    return Path(pth).absolute()


def get_lif_file(pth: str) -> LifFile:
    return LifFile(return_path(pth))


def get_image(pth: Union[str, PathLike]) -> np.array:
    """ Uses PIL to load an image as an array in grayscale ('L' mode)."""
    with Image.open(return_path(pth)) as im:
        im = im.convert(mode="L")
        return np.array(im)


def lif_read(pth: str, debug: str = '') -> list[FRCSet]:
    """
    Read a LIF file. Assumes the scale is in pixels per um and is equal in all directions.
    If debug is True, it will only get the first image and channel.
    """
    images = []
    lif_file = get_lif_file(pth)
    imgs = lif_file.get_iter_image()
    if debug == 'single':
        imgs = [next(imgs)]
    elif debug == 'two_set':
        imgs_itered = [img for img in imgs]
        imgs = [imgs_itered[0], imgs_itered[-1]]
    for img in imgs:
        pixels_per_um = img.scale[0]
        name = img.name
        um_per_pixel = 1 / pixels_per_um
        nm_per_pixel = um_per_pixel * 1000
        NA = img.settings["NumericalAperture"] if "NumericalAperture" in img.settings else None
        lam = img.settings["StedDelayWavelength"] if "StedDelayWavelength" in img.settings else None
        measurements = []
        channels = img.get_iter_c(t=0, z=0)
        if debug == 'single':
            channels = [next(channels)]
        elif debug == 'two_set':
            channels_itered = [channel for channel in channels]
            channels = [channels_itered[0], channels_itered[-1]]
        for i, c in enumerate(channels):
            dip_im = np.array(c)
            settings = FRCMeasureSettings(NA=NA, lambda_excite_nm=lam, nm_per_pixel=nm_per_pixel)
            measurement = FRCMeasurement(image=dip_im, group_name=name, index=i, settings=settings)
            measurements.append(measurement)
        frc_image = FRCSet(name, measurements)
        images.append(frc_image)
    return images


def image_read(pth: str, pth2: Optional[str] = None):
    img1 = get_image(pth)
    if pth2 is not None:
        img2 = get_image(pth2)
        return frc2_set(img1, img2)
    else:
        return frc1_set(img1)



import pathlib as path
import time


__all__ = ['create_save']


def create_save(folder_path: str = './results/', folder_name: str = 'frc_plots', add_timestamp=True):
    folder_path = folder_path.replace('\\', '/').removesuffix('/')
    if add_timestamp:
        timestamp = int(time.time())
        path_str = f"{folder_path}/{timestamp}-"
    else:
        path_str = f"{folder_path}/"
    path_result = path.Path(f"{path_str}{folder_name}")
    path_result.mkdir(parents=True, exist_ok=True)
    return path_result



from typing import Union, Optional

from dataclasses import dataclass, field

import numpy as np


__all__ = ['Curve',  'CurveTask', 'FRCSet', 'FRCMeasurement', 'FRCMeasureSettings', 'frc1_set', 'frc2_set', 'frc_set', 'frc_measure']


@dataclass
class FRCMeasureSettings:
    nm_per_pixel: float
    extra: dict = field(default_factory=dict)
    smooth_curve: bool = False
    NA: Optional[float] = None
    lambda_excite_nm: Optional[float] = None


@dataclass
class Curve:
    key: str

    curve_x: np.ndarray
    curve_y: np.ndarray
    frc_res: float

    curve_title: str
    curve_label: str
    desc: str

    res_sd: float
    res_y: float
    thres: np.ndarray
    thres_name: str

    measure: 'FRCMeasurement'


@dataclass
class FRCMeasurement:
    """
    An FRCMeasurement represents a single measurement within an FRCSet and contains the actual data as
    a DIP Image (which itself can be converted to a numpy array at no cost). If only one image is provided,
    it will calculate a 1FRC. If two are provided, it will calculate the standard FRC between 2 images.
    """
    group_name: str
    index: int
    settings: FRCMeasureSettings
    image: np.ndarray
    curve_tasks: Optional[list] = None
    image_2: Optional[np.ndarray] = None
    curves: list[Curve] = None
    id: str = None

    def __post_init__(self):
        self.update_id()

    def update_id(self, index: Optional[int] = None):
        if index is not None:
            self.index = index
        self.id = f"{self.group_name}-c{self.index}"
        return self


class FRCSet:
    """
    The FRCSet represents a single measurement group, where each measurement in the group has the same
    image dimensions, is square and is windowed to prevent FFT artifacts. I.e. they are fully preprocessed.
    Measurements should also have occurred under similar conditions, i.e. same camera, lens, with only
    some properties variable per image.
    """
    name: str
    measurements: list[FRCMeasurement]

    def __init__(self, name, measurements):
        self.name = name
        self.measurements = measurements


@dataclass
class CurveTask:
    key: str = ''
    method: str = '1FRC1'
    smooth: bool = False
    smooth_frac: float = 0.2
    avg_n: int = None
    threshold: str = '1/7'

    def __post_init__(self):
        if self.avg_n is None:
            if self.method == '2FRC' or self.smooth:
                self.avg_n = 1
            else:
                self.avg_n = 5





def frc_measure(img: np.ndarray, img_2: Optional[np.ndarray] = None, set_name='FRCmeasure', nm_per_pixel=1, **kwargs):
    settings = FRCMeasureSettings(nm_per_pixel, **kwargs)
    return FRCMeasurement(set_name, 0, settings, img, image_2=img_2)


def frc_set(measure: FRCMeasurement, *args: FRCMeasurement, name='FRCset'):
    measures = (measure, *args)
    indexed_measures = [measure.update_id(i) for i, measure in enumerate(measures)]
    return FRCSet(name, indexed_measures)


def frc1_set(img: np.ndarray, name='1FRCset', nm_per_pixel=1, **kwargs):
    settings = FRCMeasureSettings(nm_per_pixel=nm_per_pixel, **kwargs)
    measurement = FRCMeasurement(name, 0, settings, img)
    return FRCSet(name, [measurement])


def frc2_set(img1: np.ndarray, img2: np.ndarray, name='2FRCset', nm_per_pixel=1, **kwargs):
    settings = FRCMeasureSettings(nm_per_pixel=nm_per_pixel, **kwargs)
    measurement = FRCMeasurement(name, 0, settings, img1, image_2=img2)
    return FRCSet(name, [measurement])



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['analyzefrc']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.2,<9.0.0',
 'deco>=0.6.3,<0.7.0',
 'frc>=0.1.0,<0.2.0',
 'loess>=2.1.1,<3.0.0',
 'matplotlib>=3.4,<4.0',
 'readlif>=0.6.5,<0.7.0']

setup_kwargs = {
    'name': 'analyzefrc',
    'version': '0.1.0',
    'description': 'Plots, analysis and resolution measurement of microscopy images using Fourier Ring Correlation (FRC).',
    'long_description': None,
    'author': 'Tip ten Brink',
    'author_email': 'T.M.tenBrink@student.tudelft.nl',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



from typing import Optional, Union, Callable

from dataclasses import dataclass
from frc.deps_types import NoIntersectionException
import frc.frc_functions as frcf
import frc.utility as util
import numpy as np
from deco import concurrent, synchronized
from loess.loess_1d import loess_1d

from analyzefrc.read import FRCMeasureSettings, FRCMeasurement, FRCSet, Curve, CurveTask


__all__ = ['group_all', 'group_sets', 'group_measures', 'process_frc']


@dataclass
class ProcessTask:
    processings: list[Callable]
    measure: FRCMeasurement


def process_task(task: ProcessTask):
    print(f"{task.measure.group_name} {task.measure.index}")
    for processing in task.processings:
        task.measure = processing(task.measure)
    return task.measure.curves


@concurrent
def process_task_conc(task: ProcessTask):
    for processing in task.processings:
        task.measure = processing(task.measure)
    return task.measure.curves


@synchronized
def process_measures_conc(measure_tasks: list[ProcessTask]) -> dict:
    processed_measures = {}
    for task in measure_tasks:
        processed_measures[(task.measure.group_name, task.measure.index)] = process_task_conc(task)
    return processed_measures


def process_measures(measure_tasks: list[ProcessTask]) -> dict:
    processed_measures = {}
    for task in measure_tasks:
        processed_measures[(task.measure.group_name, task.measure.index)] = process_task(task)

    return processed_measures


def create_tasks(frc_sets: Union[list[FRCSet], FRCSet], preprocess=True) -> list[ProcessTask]:
    if isinstance(frc_sets, FRCSet):
        frc_sets = [frc_sets]
    process_tasks = []
    for frc_set in frc_sets:
        for measure in frc_set.measurements:
            if preprocess:
                tasks = [preprocess_measure]
            else:
                tasks = []
            tasks.append(measure_curve)
            process_tasks.append(ProcessTask(tasks, measure))

    return process_tasks


def group_all(process_name: str, curves: list[Curve]) -> dict[str, list[Curve]]:
    return {process_name: curves}


def group_measures(curves: list[Curve]) -> dict[str, list[Curve]]:
    measure_dict = {}
    for c in curves:
        if c.measure.id not in measure_dict:
            measure_dict[c.measure.id] = [c]
        else:
            measure_dict[c.measure.id] += c
    return measure_dict


def group_sets(curves: list[Curve]) -> dict[str, list[Curve]]:
    group_dict = {}
    for c in curves:
        if c.measure.group_name not in group_dict:
            group_dict[c.measure.group_name] = [c]
        else:
            group_dict[c.measure.group_name].append(c)
    return group_dict


def process_frc(process_name: str, frc_sets: Union[list[FRCSet], FRCSet], preprocess=True, concurrency=True,
                grouping: str = 'measures') -> dict[str, list[Curve]]:
    tasks = create_tasks(frc_sets, preprocess)
    processed_measures = process_measures_conc(tasks) if concurrency else process_measures(tasks)
    processed_curves = [curve for curves in processed_measures.values() for curve in curves]
    if grouping == 'sets':
        return group_sets(processed_curves)
    elif grouping == 'measures':
        return group_measures(processed_curves)
    else:
        return group_all(process_name, processed_curves)


def preprocess_measure(measure: FRCMeasurement):
    measure.image = preprocess_img(measure.image)
    if measure.image_2 is not None:
        measure.image_2 = preprocess_img(measure.image_2)
    return measure


def preprocess_img(img: np.ndarray):
    img = util.square_image(img, add_padding=False)
    return util.apply_tukey(img)


def measure_curve(measure: FRCMeasurement):
    img = measure.image
    # Can be None
    img2 = measure.image_2
    img_size = img.shape[0]
    nm_per_pixel = measure.settings.nm_per_pixel
    frc_curve = frcf.one_frc(img, 1)
    xs_pix = np.arange(len(frc_curve)) / img_size
    xs_nm_freq = xs_pix * (1 / nm_per_pixel)

    if measure.curve_tasks is None:
        if measure.image_2 is None:
            measure.curve_tasks = [CurveTask(key='curve1')]
        else:
            measure.curve_tasks = [CurveTask(key='curve1', method='2FRC', avg_n=1)]

    curves = []
    for curve_i, curve_task in enumerate(measure.curve_tasks):

        if curve_task.method == '1FRC' or curve_task.method == '1FRC1':
            def frc_func(img_frc, _img2):
                return frcf.one_frc(img_frc, 1)
        elif curve_task.method == '2FRC':
            def frc_func(img_frc_1, img_frc_2):
                return frcf.two_frc(img_frc_1, img_frc_2)
        else:
            raise ValueError("Unknown method {}".format(curve_task.method))
        frc_curve = frc_func(img, img2)
        frc_curves = [frc_curve]
        for i in range(curve_task.avg_n - 1):
            calculated_frc = frc_func(img, img2)
            frc_curve += calculated_frc
            frc_curves.append(calculated_frc)

        frc_curve /= curve_task.avg_n

        label = curve_task.method

        smooth_desc = ""
        if curve_task.smooth:
            smooth_desc = " LOESS smoothing (point frac: {}).".format(curve_task.smooth_frac)
            xs_pix, frc_curve, wout = loess_1d(xs_pix, frc_curve, frac=curve_task.smooth_frac)

        frc_res = -1
        sd_str = " "

        res_sd = 0
        res_y = 0
        thres_curve = None

        try:
            frc_res, res_y, thres = frcf.frc_res(xs_nm_freq, frc_curve, img_size, threshold=curve_task.threshold)
            thres_curve = thres(xs_pix)
            frc_res_list = []
            for res_curve in frc_curves:
                try:
                    frc_res, _, _ = frcf.frc_res(xs_nm_freq, res_curve, img_size, threshold=curve_task.threshold)
                    frc_res_list.append(frc_res)
                except NoIntersectionException:
                    pass
            res_sd = np.std(np.array(frc_res_list))
            if res_sd > 0:
                sd_str = "± {:.2g} ".format(res_sd)

        except NoIntersectionException as e:
            print(e)
        avg_desc = f"{curve_task.avg_n} curves averaged." if curve_task.avg_n > 1 else ""
        desc = "Resolution ({} threshold): {:.3g} {}nm. {} {}".format(curve_task.threshold, frc_res, sd_str,
                                                                      smooth_desc,
                                                                      avg_desc)

        curve = Curve(f"{measure.group_name}-{measure.index}-{curve_i}", xs_nm_freq, frc_curve, frc_res,
                      f"{measure.group_name} measurement {measure.index}", label, desc,
                      res_sd, res_y, thres_curve, curve_task.threshold, measure)
        curves.append(curve)
    measure.curves = curves
    return measure



from typing import Union, Callable

import time
import pathlib as path
from pathlib import Path

import matplotlib.pyplot as plt

from analyzefrc.read import Curve


__all__ = ['CurvePlot', 'plot_all']


class CurvePlot:
    title: str
    curves: list[Curve]

    def __init__(self, *args: Union[Curve, list[Curve]], title='Title'):
        self.title = title
        curves = []
        for arg in args:
            if isinstance(arg, list):
                curves += arg
            elif isinstance(arg, Curve):
                curves.append(arg)
            else:
                raise ValueError("Arguments must be Curves or lists of Curves!")
        self.curves = curves

    def plot(self, show=False, save=False, save_directory: Path = None, ax_fig_ops: Callable = None, dpi=180, **kwargs):
        min_y = -0.1
        max_y = 1.1

        fig, ax = plt.subplots()

        descs = []
        for curve in self.curves:
            ax.plot(curve.curve_x, curve.curve_y, label=curve.curve_label, zorder=2)

            ax.plot(curve.curve_x, curve.thres, c='darkgoldenrod', label=f"Threshold curve ({curve.thres_name})",
                    zorder=1)
            ax.vlines(1 / curve.frc_res, min_y, curve.res_y, ls='dashed', colors=['darkgoldenrod'], zorder=3)
            descs.append(curve.desc)

        ax.set_xlabel("Spatial frequency ($\\mathrm{nm}^{-1}$)")
        ax.set_ylabel("FRC")

        ax.legend()
        ax.set_ylim(min_y, max_y)

        ax.set_title(self.title)

        fig.supxlabel('\n'.join(descs), fontsize='small')
        fig.set_tight_layout(True)

        if ax_fig_ops is not None:
            ax, fig = ax_fig_ops(self, ax, fig)

        if save:
            if save_directory is None:
                raise ValueError("Save folder path must be given when saving!")
            save_path = save_directory.joinpath(self.title)
            i = 0
            while save_path.exists():
                save_path = save_path.parent.joinpath(f"{self.title}{i}")
                i += 1
            fig.savefig(save_path, dpi=dpi, **kwargs)
        if show:
            plt.show()
        plt.close(fig)


def plot_all(*multiple_groups: dict[str, list[Curve]], show=True, save=False, save_directory=None, dpi=180):
    curve_plots = []
    for multiple_group in multiple_groups:
        for group_name, group in multiple_group.items():
            curve_plots.append(CurvePlot(group, title=group_name))
    for curve_plot in curve_plots:
        curve_plot.plot(show=show, save=save, save_directory=save_directory, dpi=dpi)



import pathlib as path
import pickle
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit

dir_path = path.Path("../results_sav/{}".format('1636714207-1FRC1 mike3'))
sheet = pd.read_excel('../data/sted/2021_10_05_XSTED_NileRed_variation_excitation_power_laser_settings.xlsx')
pth = dir_path.joinpath("./res.pickle")

with open(pth, 'rb') as handle:
    d = pickle.load(handle)

sheet = sheet.loc[~sheet['Series'].isna()]
sheet = sheet.astype({"Series": int, "Ch": int})
sheet = sheet.astype({"Series": str, "Ch": str})
series_df = sheet.set_index(['Series', 'Ch'])

v = {v[0]: v[1] for k, v in d.items()}
res = []
res_uncertain = []
for key, value in v.items():
    spl = key.split(')')[-1].split('-')
    sers = str(int("".join([n for n in spl[0] if n.isdigit()])))
    c = str(int(spl[1].split('o')[0]))
    res.append([sers, c, value[0], value[1], value[2]])
    res_uncertain.append([sers, c, value[1]])
res_df = pd.DataFrame(res, columns=['Series', 'Ch', 'res_nm', 'res_uncertain', 'NA']).set_index(['Series', 'Ch'])
# res_uncertain_df = pd.DataFrame(res_uncertain, columns=['Series', 'Ch', 'res_uncertain']).set_index(['Series', 'Ch'])

# series_list = {series_n: sheet.loc[series_col == series_n].set_index('Ch').drop('Series', 1)
#                for series_n in [3., 6., 12.]}


# for series_n, df in series_list.items():
series_df = series_df.join(res_df, on=['Series', 'Ch'])
# series_df = series_df.join(res_uncertain_df, on=['Series', 'Ch'])
series_df['res_relative'] = pd.Series([0] * len(series_df['res_nm']))
series_df['res_relative_uncertain'] = pd.Series([0] * len(series_df['res_nm']))
for srs_n in ['3', '6', '12']:
    srs = series_df.loc[srs_n, :]
    res_base = srs.loc['1', 'res_nm']
    series_df.loc[(srs_n, slice(None)), 'res_relative'] = series_df.loc[(srs_n, slice(None)), 'res_nm'] / res_base
    series_df.loc[(srs_n, slice(None)), 'res_relative_uncertain'] = series_df.loc[(srs_n, slice(
        None)), 'res_uncertain'] / res_base
ii = series_df.index.get_level_values('Series').to_numpy()
series_arr = series_df.to_numpy().astype(float)
series_colors = [np.empty(ii.shape, dtype=tuple)]
c_map = plt.cm.get_cmap(lut=3)
color_series = []
for i, srs_n in enumerate(['3', '6', '12']):
    srs_loc = ii == srs_n
    c_map_i = c_map(i)
    color_series.append((srs_loc, c_map_i))

std_data = series_arr[:, 1]
exc_data = series_arr[:, 0]
res_data = series_arr[:, 2]
NA_data = series_arr[:, 4]
res_data_relat = series_arr[:, 5]
res_err_data_relat = series_arr[:, 6]
res_err_data = series_arr[:, 3]
print()
fig, ax = plt.subplots()


def sted_func(NA, x, b):
    return 561 / (2 * NA * np.sqrt(1 + x / b))


for srs_loc, color in color_series:
    x_data = std_data[srs_loc] * 100
    y_data = res_data[srs_loc]
    y_err = res_err_data[srs_loc]
    NA = NA_data[srs_loc][0]
    series = ii[srs_loc][0]
    # init_a = y_data[0] * 2 * NA
    popt, pcov = curve_fit(lambda x, a: sted_func(NA, x, a), x_data, y_data, bounds=([1], [50]), p0=[25], sigma=y_err)
    ax.errorbar(x_data, y_data, yerr=y_err, c=color, ls='', marker='.',
                capsize=2, ms=4, label=f"Series{series} (NA={NA})")
    xs = np.linspace(0, 100)
    ax.plot(xs, sted_func(NA, xs, popt[0]), c=color, ls='dashed')

ax.legend()
ax.set_xlabel("STED intensity (%)")
ax.set_ylabel("Resolution 1/7 threshold (nm)")
ax.set_title("2021/10/5 XSTED NileRed STED equation")
desc = "STED 775nm. 561nm excitation."
fig.supxlabel(desc, fontsize='small')
fig.set_tight_layout(True)
fig.savefig(dir_path.joinpath("./stedcurves.png"), dpi=180)
series_df.to_excel(dir_path.joinpath("./resolutions.xlsx"))
plt.show()

# for srs_loc, color in color_series:
#     ax.errorbar(std_data[srs_loc], res_data_relat[srs_loc], yerr=res_err_data_relat[srs_loc], c=color, ls='', marker='.', capsize=2, ms=4)
# plt.show()
